# -*- coding: utf-8 -*-
"""
All the recursive, argument-parsin', context-managin' goodness.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
from __future__ import print_function

import re
import sys


#: Globally-unique value used by commands to indicate they want clik to send
#: the exception (if one occurred) in addition to the child exit code in
#: response to a ``yield``.
#:
#: :type: :class:`object`
catch = object()

#: Name of the magic variable containing the parsed arguments.
#:
#: :type: :class:`str`
ARGS = 'args'

#: Name of the argument that contains the stack of commands to be run.
#:
#: :type: :class:`str`
STACK = '_clik_stack'

#: If a parent has more than this number of subcommands, the help output will
#: ``{command}`` instead of the full list of subcommands.
#:
#: :type: :class:`int`
SHOW_SUBCOMMANDS = 3


class BareAlreadyRegisteredError(Exception):
    """Raised when a bare command has already been registered."""

    def __init__(self, command):
        """
        Initialize the exception.

        :param command: Command the user is trying to register.
        :type command: :class:`Command`
        """
        fmt = 'Bare command already registered for command "%s"'
        super(BareAlreadyRegisteredError, self).__init__(fmt % command._name)

        #: Command caller was trying to register as the bare command.
        #:
        #: :type: :class:`Command`
        self.command = command


class Command(object):
    """The invisible backend driving most of what the user interacts with."""

    @staticmethod
    def _split_docstring(x=None):
        parts = list(filter(None, re.split(r'\n\s*\n', x.__doc__ or '', 1)))
        return parts + [None] * (2 - len(parts))

    def __init__(self, ctx, fn, name=None, alias=None, aliases=None):
        """
        Initialize the command object.

        :param ctx: Context object.
        :type ctx: :class:`clik.context.Context`
        :param fn: Generator function -- the actual command.
        :param str name: Name of the command or ``None``. If ``None``, name
                         will be ``fn.__name__``.
        :param str alias: Command alias or ``None``. If ``None``, the command
                          has no aliases. If this and ``aliases`` are both
                          supplied, ``alias`` will be prepended to the
                          ``aliases`` list.
        :param aliases: List of additional aliases for the command or ``None``.
        :type aliases: :class:`list` or ``None``
        """
        #: Tuple of aliases for this command.
        #:
        #: :type: :class:`tuple` of :class:`str`
        self._aliases = ()

        #: :class:`Command` instance for the bare command or ``None`` if bare
        #: command is not set.
        #:
        #: :type: :class:`Command` or ``None``
        self._bare = None

        #: List of child commands.
        #:
        #: :type: :class:`list` of :class:`Command` instances
        self._children = []

        #: Context object for this command. This context is shared between all
        #: command instances associated with a :class:`clik.app.App`.
        #:
        #: :type: :class:`clik.context.Context`
        self._ctx = ctx

        #: Generator function -- the actual command.
        #:
        #: :type: generator function
        self._fn = fn

        #: In-progress generator for :attr:`_fn`. This is the object that we
        #: call :meth:`generator.send` and :meth:`generator.next` on.
        #:
        #: :type: :class:`generator`
        self._generator = None

        #: Canonical name of the command.
        #:
        #: :type: :class:`str`
        self._name = 'Unnamed'

        #: Parent command. For the root :class:`clik.app.App` instance, this
        #: is ``None``. Set in :meth:`_configure_parser`.
        #:
        #: :type: :class:`Command` or ``None``
        self._parent = None

        #: Parser for this command. Set in :meth:`_configure_parser`.
        #:
        #: :type: :class:`clik.argparse.ArgumentParser`
        self._parser = None

        if name is None:
            self._name = fn.__name__
        else:
            self._name = name

        if aliases is None:
            aliases = []
        else:
            aliases = list(aliases)
        if alias is not None:
            aliases.insert(0, alias)
        self._aliases = tuple(aliases)

        self.__doc__ = fn.__doc__

    def bare(self, fn):
        """
        Register a bare command.

        :param fn: Generator function -- the bare command.
        :raise: :exc:`BareAlreadyRegisteredError` if a bare command has already
                been registered.
        """
        if self._bare is not None:
            raise BareAlreadyRegisteredError(self)
        self._bare = Command(self._ctx, fn)
        return self._bare

    def __call__(self, fn=None, name=None, alias=None, aliases=None):
        """
        Add subcommands to this command.

        Basic use::

            @myapp
            def mysubcommand():
                yield

        Customizing the subcommand::

            @myapp(name='subcommand', alias='sub', aliases=['s'])
            def mysubcommand():
                yield

        :param fn: Generator function or ``None``. If ``fn`` is supplied,
                   all other arguments are ignored.
        :param str name: Name of the command or ``None``.
        :param str alias: Command alias. See :meth:`__init__` for information
                          on how aliases are handled.
        :param aliases: List of additional aliases for the command or ``None``.
        :type aliases: :class:`list` or ``None``
        """
        def decorate(fn):
            self._children.append(Command(self._ctx, fn, name, alias, aliases))
            return self._children[-1]
        if fn is None:
            return decorate
        return decorate(fn)

    def _configure_parser(self, parser, parent=None, stack=None):
        self._parser = parser
        parser._clik_command = self

        if stack is None:
            stack = []
        if parent is not None:
            self._parent = parent

        self._generator = self._fn()
        with self._ctx(parser=parser):
            ec = next(self._generator)
            if ec:
                return ec

        stack = stack + [self]
        if self._bare:
            self._bare._generator = self._bare._fn()
            with parser._clik_bare_arguments():
                with self._ctx(parser=parser):
                    ec = next(self._bare._generator)
                    if ec:
                        return ec
            parser.set_defaults(**{STACK: stack + [self._bare]})

        if self._children:
            if len(self._children) > SHOW_SUBCOMMANDS:
                metavar = '{command}'
            else:
                names = ','.join([c._name for c in self._children])
                metavar = '{%s}' % names
            subparsers = parser.add_subparsers(
                dest='command',
                metavar=metavar,
                title='subcommands',
            )
            if not self._bare:
                subparsers.required = True
            for child in self._children:
                description, epilog = self._split_docstring(child._fn)
                subparser = subparsers.add_parser(
                    child._name,
                    aliases=child._aliases,
                    description=description,
                    epilog=epilog,
                    help=description,
                )
                ec = child._configure_parser(subparser, self, stack)
                if ec:
                    return ec

        if not self._bare and not self._children:
            parser.set_defaults(**{STACK: stack})

    def _check_bare_arguments(self):
        if self._parent is not None:
            parser = self._parent._parser
            if parser._clik_bare_dests is not None:
                args = self._ctx.get(ARGS)
                for dest in parser._clik_bare_dests:
                    if getattr(args, dest) != parser.get_default(dest):
                        parser.print_usage(sys.stderr)
                        for action in parser._actions:  # pragma: no branch
                            if action.dest == dest:
                                break
                        else:  # pragma: no cover (unreachable)
                            raise Exception('this code should be unreachable')
                        options = '/'.join(action.option_strings)
                        err = 'unrecognized arguments when calling subcommand'
                        fmt = '%s: error: %s: %s'
                        msg = fmt % (parser.prog, err, options)
                        print(msg, file=sys.stderr)
                        return 1
                    delattr(args, dest)

    def _run(self, child=False):
        stack = getattr(self._ctx.get(ARGS), STACK)

        if not child:
            for command in stack:
                ec = command._check_bare_arguments()
                if ec:
                    return ec

        command = stack.pop(0)

        if stack:
            def run_children():
                return stack[0]._run(child=True)

            with self._ctx(run_children=run_children):
                try:
                    ec = next(command._generator)
                except StopIteration:
                    ec = 0
            if ec and ec is not catch:
                return ec

            if stack:
                if ec is catch:
                    ec = 0
                    exception = None
                    try:
                        ec = run_children()
                    except Exception as e:
                        exception = e
                    send = (ec, exception)
                else:
                    send = ec = run_children()
                try:
                    ec = command._generator.send(send)
                except StopIteration:
                    pass
        else:
            def run_children():
                return 0

            with self._ctx(run_children=run_children):
                try:
                    ec = next(command._generator)
                except StopIteration:
                    ec = 0

        return ec



# -*- coding: utf-8 -*-
"""
Manage bindings for :class:`clik.magic.Magic` variables.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
import contextlib

from clik.compat import iteritems


class LockedMagicError(Exception):
    """Raised when trying to acquire a magvar that is already acquired."""

    def __init__(self, name):
        """
        Initialize the exception.

        :param str name: Name of the magic variable that is locked.
        """
        msg = 'The magic variable "%s" is currently locked' % name
        super(LockedMagicError, self).__init__(msg)

        #: Name of the magic variable that is locked.
        #:
        #: :type: :class:`str`
        self.name = name


class MagicNameConflictError(Exception):
    """Raised when trying to register an already-registered magvar."""

    def __init__(self, name):
        """
        Initialize the exception.

        :param str name: Name of the magic variable that is already registered.
        """
        msg = 'The magic variable name "%s" is already registered' % name
        super(MagicNameConflictError, self).__init__(msg)

        #: Name of the magic variable that is already registered.
        #:
        #: :type: :class:`str`
        self.name = name


class UnregisteredMagicNameError(Exception):
    """Raised when trying to access an unregistered magic variable."""

    def __init__(self, name):
        """
        Initialize the exception.

        :param str name: Name of the unregistered magic variable the caller
                         was attempting to access.
        """
        msg = 'The magic variable "%s" is not registered' % name
        super(UnregisteredMagicNameError, self).__init__(msg)

        #: Name of the magic variable that is unregistered.
        #:
        #: :type: :class:`str`
        self.name = name


class UnboundMagicError(Exception):
    """Raised when trying to access a magic variable that is not bound."""

    def __init__(self, name):
        """
        Initialize the exception.

        :param str name: Name of the unbound magic variable the caller was
                         attempting to access.
        """
        msg = 'The magic variable "%s" is not bound' % name
        super(UnboundMagicError, self).__init__(msg)

        #: Name of the magic variable that is unbound.
        #:
        #: :type: :class:`str`
        self.name = name


class Context(object):
    """Bindings manager for magic variables."""

    def __init__(self):
        """Initialize the context."""
        self._registry = []
        self._state = {}

    @contextlib.contextmanager
    def __call__(self, **kwargs):
        """
        Context manager for :meth:`push` -ing ``kwargs`` during a code block.

        Example::

           context = Context()
           context.register('foo')
           with context(foo='bar'):
               pass  # do some stuff

        Before the block, each key/value pair in ``kwargs`` is passed to
        :meth:`push`. After the block, each key is :meth:`pop` -ped.
        """
        keys = []
        for key, value in iteritems(kwargs):
            self.push(key, value)
            keys.append(key)
        yield
        for key in keys:
            self.pop(key)

    @contextlib.contextmanager
    def acquire(self, *magic_variables):
        """
        Context manager to lock ``magic_variables`` from use by other contexts.

        Only one context at a time is allowed to control the binding of a magic
        variable. Using this context manager ensures the caller can safely
        manipulate the binding without interference from other contexts::

           foo = Magic('foo')
           context1 = Context()
           context1.register('foo')
           context2 = Context()
           context2.register('foo')
           with context1.acquire(foo):
               with context1(foo='bar'):
                   pass  # do some stuff
               with context2.acquire(foo):
                   # BOOM! LockedMagicError gets thrown

        Before the block, each of the ``magic_variables`` is checked to see if
        it currently has a context. If so, :exc:`LockedMagicError` is thrown.
        Otherwise, the context is set to this instance.

        After the block, the context for each magic variable is reset to
        ``None``, freeing it up for use by other contexts.

        :raise: :exc:`LockedMagicError` if one of ``magic_variables`` is
                already acquired
        """
        for variable in magic_variables:
            if variable._Magic__context is not None:
                raise LockedMagicError(variable._Magic__context)
            self.register(variable._Magic__name)
            object.__setattr__(variable, '_Magic__context', self)
        try:
            yield
        finally:
            for variable in magic_variables:
                self.unregister(variable._Magic__name)
                object.__setattr__(variable, '_Magic__context', None)

    def _assert_in_registry(self, name):
        if name not in self._registry:
            raise UnregisteredMagicNameError(name)

    def get(self, name):
        """
        Return currently-bound value of magic variable named ``name``.

        :param str name: Name of magic variable.
        :raise: :exc:`UnregisteredMagicNameError` if ``name`` is not registered
        :raise: :exc:`UnboundMagicError` if variable is not currently bound
        """
        self._assert_in_registry(name)
        if not self._state[name]:
            raise UnboundMagicError(name)
        return self._state[name][0]

    def push(self, name, obj):
        """
        Push a value on to a variable's stack, rebinding its current value.

        :param str name: Name of magic variable.
        :param obj: New value to push on to the stack.
        :raise: :exc:`UnregisteredMagicNameError` if ``name`` is not registered
        """
        self._assert_in_registry(name)
        self._state[name].insert(0, obj)

    def pop(self, name):
        """
        Pop and return the current value off the variable's stack.

        This rebinds the variable to the next-highest item on the stack.

        :param str name: Name of magic variable.
        :raise: :exc:`UnregisteredMagicNameError` if ``name`` is not registered
        :raise: :exc:`UnboundMagicError` if variable is not currently bound
        """
        self._assert_in_registry(name)
        if not self._state[name]:
            raise UnboundMagicError(name)
        return self._state[name].pop(0)

    def register(self, name):
        """
        Register a magic variable name.

        Requiring registration prevents accidental conflicts between modules.
        If two modules (which may not know about each other) both try to
        register the same magic variable, clik will thrown an exception.

        :param str name: Name of magic variable.
        :raise: :exc:`MagicNameConflictError` if ``name`` is already registered
        """
        if name in self._registry:
            raise MagicNameConflictError(name)
        self._registry.append(name)
        self._state[name] = []

    def unregister(self, name):
        """
        Unregister a magic variable name.

        :param str name: Name of magic variable.
        :raise: :exc:`UnregisteredMagicNameError` if ``name`` is not registered
        """
        self._assert_in_registry(name)
        self._registry.remove(name)
        del self._state[name]



# -*- coding: utf-8 -*-
"""
Most of the hackery that makes clik tick.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
from __future__ import absolute_import, print_function

import argparse
import contextlib
import sys

from clik.compat import PY2, PY26, PY33


#: Name of the argument that contains whether to allow unknown
#: arguments.
#:
#: :type: :class:`str`
ALLOW_UNKNOWN = '_clik_unknown'


class ArgumentParserExit(Exception):
    """Raised instead of allowing :mod:`argparse` to call :func:`sys.exit`."""

    def __init__(self, code):
        """
        Initialize the exception.

        :param int code: Exit code.
        """
        fmt = 'argument parser exited with return code %i'
        super(ArgumentParserExit, self).__init__(fmt % code)

        #: Exit code.
        #:
        #: :type: :class:`int`
        self.code = code


class BareUnsupportedFeatureError(Exception):
    """Raised when using a feature that is not supported by bare commands."""

    def __init__(self, feature):
        """
        Initialize the exception.

        :param str feature: Name of the feature that is unsupported.
        """
        msg = 'Feature is not supported for bare commands: %s' % feature
        super(BareUnsupportedFeatureError, self).__init__(msg)

        #: Feature that is unsupported.
        #:
        #: :type: :class:`str`
        self.feature = feature


class UnknownArgsUnsupportedError(Exception):
    """Raised when trying to allow unknown args on a command with children."""

    def __init__(self):
        """Initialize the exception."""
        msg = 'Unknown args are not allowed for commands with subcommands'
        super(UnknownArgsUnsupportedError, self).__init__(msg)


class HelpFormatter(argparse.HelpFormatter):
    """Format usage with no trailing newline on usage."""

    def _format_usage(self, *args, **kwargs):
        parent = super(HelpFormatter, self)
        return parent._format_usage(*args, **kwargs)[:-1]


class ArgumentParser(argparse.ArgumentParser):
    """:class:`argparse.ArgumentParser` specialized for clik."""

    def __init__(self, *args, **kwargs):
        """Initialize -- same arguments as :class:`argparse.ArgumentParser`."""
        self._clik_bare_command_mode = False
        self._clik_bare_dests = None
        self._clik_command = None
        kwargs.setdefault('formatter_class', HelpFormatter)
        super(ArgumentParser, self).__init__(*args, **kwargs)

    def exit(self, status=0, message=None):
        """
        Override default behavior to avoid interpreter exit.

        By default, the parser calls :func:`sys.exit`. In certain situations --
        namely testing -- we don't actually want to exit the Python
        interpreter.

        So instead of exiting, this throws an :exc:`ArgumentParserExit`
        exception which can be caught by the caller.

        :param int status: Exit status.
        :param str message: Optional message. If supplied, will be printed
                            to :data:`sys.stderr` before raising the exception.
        :raise: :exc:`ArgumentParserExit`
        """
        if message:
            print(message, end='', file=sys.stderr)
        raise ArgumentParserExit(status)

    @contextlib.contextmanager
    def _clik_bare_arguments(self):
        """
        Context manager for bare command mode.

        When the parser is in bare command mode, it disallows certain features
        (like positional args and mutually exclusive groups). In addition, the
        argument destinations are recorded in order to do some post-processing
        before running the selected command.
        """
        self._clik_bare_dests = []
        self._clik_bare_command_mode = True
        yield
        self._clik_bare_command_mode = False

    def add_argument(self, *args, **kwargs):
        """
        Override default behavior to disallow posargs in bare commands.

        :raise: :class:`BareUnsupportedFeatureError` if adding a positional
                argument to a bare command.
        """
        argument = super(ArgumentParser, self).add_argument(*args, **kwargs)
        if self._clik_bare_command_mode:
            if argument.nargs:
                msg = 'positional arguments (dest: %s)' % argument.dest
                raise BareUnsupportedFeatureError(msg)
            self._clik_bare_dests.append(argument.dest)
        return argument

    def add_mutually_exclusive_group(self, *args, **kwargs):
        """
        Override default behavior to disallow mutex groups in bare commands.

        :raise: :class:`BareUnsupportedFeatureError` if adding a mutually
                exclusive group to a bare command.
        """
        if self._clik_bare_command_mode:
            raise BareUnsupportedFeatureError('mutually exclusive groups')
        s = super(ArgumentParser, self)
        return s.add_mutually_exclusive_group(*args, **kwargs)

    def allow_unknown_args(self):
        """
        Allow unknown arguments, putting them in ``clik.unknown_args``.

        :raise: :exc:`UnknownArgsUnsupportedError` if this parser has
                subparsers -- unknown arguments are allowed *only* on
                leaf commands.
        """
        if self._clik_command._children:
            raise UnknownArgsUnsupportedError
        self.set_defaults(**{ALLOW_UNKNOWN: True})

    def _clik_format_usage(self, formatter):
        bare_dests = self._clik_bare_dests
        actions, bare_actions, subparsers = [], [], None
        for action in self._actions:
            if isinstance(action, argparse._SubParsersAction):
                subparsers = action
                actions.append(action)
            else:
                bare_actions.append(action)
                if bare_dests is None or action.dest not in bare_dests:
                    actions.append(action)
        prefix = 'usage: '
        mutex_groups = self._mutually_exclusive_groups
        if bare_dests is None or subparsers:
            formatter.add_usage(self.usage, actions, mutex_groups, prefix)
            prefix = '       '
        if bare_dests is not None:
            formatter.add_usage(self.usage, bare_actions, mutex_groups, prefix)
        return formatter

    def format_usage(self):
        """Override default behavior to use clik's formatter."""
        return self._clik_format_usage(self._get_formatter()).format_help()

    def format_help(self):
        """Override default behavior to support formatting bare commands."""
        formatter = self._get_formatter()
        self._clik_format_usage(formatter)
        formatter.add_text('\n')
        formatter.add_text(self.description)
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()


if PY33 or PY26:  # pragma: no cover (~copypaste of Python 3.4 implementation)
    # LINT: Ignore undocumented function violation. Compatibility code is not
    #       formally documented.
    def __call__(self, parser, namespace, values, option_string=None):  # noqa
        parser_name = values[0]
        arg_strings = values[1:]

        if self.dest is not argparse.SUPPRESS:
            setattr(namespace, self.dest, parser_name)

        try:
            parser = self._name_parser_map[parser_name]
        except KeyError:
            fmt = 'unknown parser %(parser_name)r (choices: %(choices)s)'
            args = {
                'parser_name': parser_name,
                'choices': ', '.join(self._name_parser_map),
            }
            raise argparse.ArgumentError(self, fmt % args)

        subnamespace, arg_strings = parser.parse_known_args(arg_strings, None)
        for key, value in vars(subnamespace).items():
            setattr(namespace, key, value)

        if arg_strings:
            vars(namespace).setdefault(argparse._UNRECOGNIZED_ARGS_ATTR, [])
            attr = getattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR)
            attr.extend(arg_strings)

    argparse._SubParsersAction.__call__ = __call__


if PY2:
    original_error = ArgumentParser.error

    # LINT: Ignore undocumented function violation. Compatibility code is not
    #       formally documented.
    def error(self, message=None):  # noqa: D103
        if message == 'too few arguments':
            for action in self._actions:
                if isinstance(action, argparse._SubParsersAction):
                    name = action.metavar
                    break
            else:
                for action in self._positionals._actions:  # pragma: no branch
                    if action.required:
                        name = action.dest
                        break
                else:  # pragma: no cover (unreachable)
                    raise Exception('this code should be unreachable')
            if action.required:
                self.print_usage(sys.stderr)
                fmt = '%s: error: the following arguments are required: %s\n'
                return self.exit(2, fmt % (self.prog, name))
        else:
            return original_error(self, message)

    ArgumentParser.error = error

    # https://gist.github.com/sampsyo/471779

    # LINT: Ignore undocumented function violation. Compatibility code is not
    #       formally documented.
    class _AliasedSubParsersPseudoAction(argparse.Action):  # noqa: D103
        def __init__(self, name, aliases, help):
            dest = name
            if aliases:
                dest += ' (%s)' % ', '.join(aliases)
            parent = super(_AliasedSubParsersPseudoAction, self)
            parent.__init__(option_strings=[], dest=dest, help=help)

    original_add_parser = argparse._SubParsersAction.add_parser

    # LINT: Ignore undocumented function violation. Compatibility code is not
    #       formally documented.
    def add_parser(self, name, **kwargs):  # noqa: D103
        aliases = kwargs.pop('aliases')
        parser = original_add_parser(self, name, **kwargs)
        for alias in aliases:
            self._name_parser_map[alias] = parser
        help = kwargs.pop('help')
        self._choices_actions.pop()
        action = _AliasedSubParsersPseudoAction(name, aliases, help)
        self._choices_actions.append(action)
        return parser

    argparse._SubParsersAction.add_parser = add_parser



# -*- coding: utf-8 -*-
"""
The ever-present utilities module.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""


class AttributeDict(dict):
    """
    Simple :class:`dict` wrapper that allows key access via attribute.

    Example::

        d = AttributeDict(foo='bar', baz='qux')
        d['foo']      # 'bar'
        d.foo         # 'bar'
        d['baz']      # 'qux'
        d.baz         # 'qux'
        d.foo = 'bup'
        d['foo']      # 'bup'
        d.foo         # 'bup'
        del d.foo
        d.foo         # KeyError

    """

    def __getattr__(self, name):
        """Get via attribute name."""
        return self[name]

    def __setattr__(self, name, value):
        """Set via attribute name."""
        self[name] = value

    def __delattr__(self, name):
        """Delete via attribute name."""
        del self[name]



# -*- coding: utf-8 -*-
"""
The command line interface kit.

Clik is a tool for writing complex command-line interfaces with minimal
boilerplate and bookkeeping.

This top-level package pulls together the end user API from the various
modules within clik.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
#: Current version.
#:
#: :type: :class:`str`
__version__ = '0.92.4'


# LINT: Ignore unused import violations. This module isn't meant to "do"
#       anything, just to make clik's API available to the end user.
from clik.app import app, args, current_app, g, parser  # noqa: F401
from clik.app import run_children, unknown_args  # noqa: F401
from clik.command import catch  # noqa: F401



# -*- coding: utf-8 -*-
"""
Python compatibility helpers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
import sys


#: Indicates whether we are on Python 2.
#:
#: :type: :class:`bool`
PY2 = sys.version_info[0] == 2

#: Indicates whether we are on Python 2.6.
#:
#: :type: :class:`bool`
PY26 = sys.version_info[0:2] == (2, 6)

#: Indicates whether we are on Python 3.3.
#:
#: :type: :class:`bool`
PY33 = sys.version_info[0:2] == (3, 3)

# LINT: Ignore undocumented function violations. Compatibility code is not
#       formally documented.

if PY2:
    def iteritems(d, *args, **kwargs):  # noqa: D103
        return d.iteritems(*args, **kwargs)

    def implements_bool(cls):  # noqa: D103
        cls.__nonzero__ = cls.__bool__
        del cls.__bool__
        return cls
else:
    def iteritems(d, *args, **kwargs):  # noqa: D103
        return iter(d.items(*args, **kwargs))

    def implements_bool(cls):  # noqa: D103
        return cls



# -*- coding: utf-8 -*-
"""
Slightly adapted version of Werkzeug's ``werkzeug.local.LocalProxy``.

Original code licensed from the Werkzeug Team under the following terms:

    Copyright (c) 2014 by the Werkzeug Team, see AUTHORS for more details.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.

        * Redistributions in binary form must reproduce the above
          copyright notice, this list of conditions and the following
          disclaimer in the documentation and/or other materials provided
          with the distribution.

        * The names of the contributors may not be used to endorse or
          promote products derived from this software without specific
          prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

:copyright: Copyright (c) 2014 by the Werkzeug Team
:authors: See Werkzeug's AUTHORS file
:license: BSD
"""
# flake8: noqa
import copy

from clik.compat import implements_bool, PY2


@implements_bool
class Magic(object):  # pragma: no cover
    def __init__(self, name):
        object.__setattr__(self, '_Magic__context', None)
        object.__setattr__(self, '_Magic__name', name)

    def _get_current_object(self):
        """Return the current object."""
        return self.__context.get(self.__name)

    @property
    def __dict__(self):
        try:
            return self._get_current_object().__dict__
        except RuntimeError:
            raise AttributeError('__dict__')

    def __repr__(self):
        try:
            obj = self._get_current_object()
        except RuntimeError:
            return '<%s unbound>' % self.__class__.__name__
        return repr(obj)

    def __bool__(self):
        try:
            return bool(self._get_current_object())
        except RuntimeError:
            return False

    def __unicode__(self):
        try:
            return unicode(self._get_current_object())
        except RuntimeError:
            return repr(self)

    def __dir__(self):
        try:
            return dir(self._get_current_object())
        except RuntimeError:
            return []

    def __getattr__(self, name):
        if name == '__members__':
            return dir(self._get_current_object())
        return getattr(self._get_current_object(), name)

    def __setitem__(self, key, value):
        self._get_current_object()[key] = value

    def __delitem__(self, key):
        del self._get_current_object()[key]

    if PY2:
        __getslice__ = lambda x, i, j: x._get_current_object()[i:j]

        def __setslice__(self, i, j, seq):
            self._get_current_object()[i:j] = seq

        def __delslice__(self, i, j):
            del self._get_current_object()[i:j]

    def __setattr__(x, n, v): return setattr(x._get_current_object(), n, v)
    def __delattr__(x, n): return delattr(x._get_current_object(), n)
    def __str__(x): return str(x._get_current_object())
    def __lt__(x, o): return x._get_current_object() < o
    def __le__(x, o): return x._get_current_object() <= o
    def __eq__(x, o): return x._get_current_object() == o
    def __ne__(x, o): return x._get_current_object() != o
    def __gt__(x, o): return x._get_current_object() > o
    def __ge__(x, o): return x._get_current_object() >= o
    def __cmp__(x, o): return cmp(x._get_current_object(), o)
    def __hash__(x): return hash(x._get_current_object())
    def __call__(x, *a, **kw): return x._get_current_object()(*a, **kw)
    def __len__(x): return len(x._get_current_object())
    def __getitem__(x, i): return x._get_current_object()[i]
    def __iter__(x): return iter(x._get_current_object())
    def __contains__(x, i): return i in x._get_current_object()
    def __add__(x, o): return x._get_current_object() + o
    def __sub__(x, o): return x._get_current_object() - o
    def __mul__(x, o): return x._get_current_object() * o
    def __floordiv__(x, o): return x._get_current_object() // o
    def __mod__(x, o): return x._get_current_object() % o
    def __divmod__(x, o): return x._get_current_object().__divmod__(o)
    def __pow__(x, o): return x._get_current_object() ** o
    def __lshift__(x, o): return x._get_current_object() << o
    def __rshift__(x, o): return x._get_current_object() >> o
    def __and__(x, o): return x._get_current_object() & o
    def __xor__(x, o): return x._get_current_object() ^ o
    def __or__(x, o): return x._get_current_object() | o
    def __div__(x, o): return x._get_current_object().__div__(o)
    def __truediv__(x, o): return x._get_current_object().__truediv__(o)
    def __neg__(x): return -(x._get_current_object())
    def __pos__(x): return +(x._get_current_object())
    def __abs__(x): return abs(x._get_current_object())
    def __invert__(x): return ~(x._get_current_object())
    def __complex__(x): return complex(x._get_current_object())
    def __int__(x): return int(x._get_current_object())
    def __long__(x): return long(x._get_current_object())
    def __float__(x): return float(x._get_current_object())
    def __oct__(x): return oct(x._get_current_object())
    def __hex__(x): return hex(x._get_current_object())
    def __index__(x): return x._get_current_object().__index__()
    def __coerce__(x, o): return x._get_current_object().__coerce__(x, o)
    def __enter__(x): return x._get_current_object().__enter__()
    def __exit__(x, *a, **kw):
        return x._get_current_object().__exit__(*a, **kw)
    def __radd__(x, o): return o + x._get_current_object()
    def __rsub__(x, o): return o - x._get_current_object()
    def __rmul__(x, o): return o * x._get_current_object()
    def __rdiv__(x, o): return o / x._get_current_object()
    if PY2:
        def __rtruediv__(x, o): return x._get_current_object().__rtruediv__(o)
    else:
        __rtruediv__ = __rdiv__
    def __rfloordiv__(x, o): return o // x._get_current_object()
    def __rmod__(x, o): return o % x._get_current_object()
    def __rdivmod__(x, o): return x._get_current_object().__rdivmod__(o)
    def __copy__(x): return copy.copy(x._get_current_object())
    def __deepcopy__(x, memo):
        return copy.deepcopy(x._get_current_object(), memo)



# -*- coding: utf-8 -*-
"""
Top-level :class:`App` class and helpers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2009-2019.
:license: BSD
"""
import sys

from clik.argparse import ALLOW_UNKNOWN, ArgumentParser, ArgumentParserExit
from clik.command import Command
from clik.context import Context
from clik.magic import Magic
from clik.util import AttributeDict


#: Magic variable containing parsed arguments.
#:
#: :type: :class:`clik.magic.Magic` -- :class:`argparse.Namespace`
args = Magic('args')

#: Magic variable containing the active application instance.
#:
#: :type: :class:`clik.magic.Magic` -- :class:`clik.app.App`
current_app = Magic('current_app')

#: Magic variable containing globals.
#:
#: :type: :class:`clik.magic.Magic` -- :class:`clik.app.AttributeDict`
g = Magic('g')

#: Magic variable containing the current parser.
#:
#: :type: :class:`clik.magic.Magic` -- :class:`clik.argparse.ArgumentParser`
parser = Magic('parser')

#: Magic variable containing the function to run child commands.
#:
#: :type: :class:`clik.magic.Magic`
run_children = Magic('run_children')

#: Magic variable containing unknown arguments.
#:
#: :type: :class:`clik.magic.Magic` -- :class:`list`
unknown_args = Magic('unknown_args')


def app(fn=None, name=None):
    """
    Decorate the main application generator function.

    If the decorator is given no arguments, the name of the application is the
    name of the decorated generator function::

        # Application will be named 'myapp' in this case
        @app
        def myapp():
            yield

    The application name can be set by passing a string to ``name``::

        # Application will be named 'theapp' in this case
        @app(name='theapp')
        def myapp():
            yield

    :param fn: Main application generator function. Name of the application
               will be set to ``fn.__name__``.
    :param name: Overrides name of application. Must not be used with the
                 ``fn`` argument (if used with ``fn``, ``name`` is ignored).
    :return: :class:`App` if ``fn`` is ``None``, otherwise a decorator
             returning :class:`App`.
    """
    def decorate(fn):
        return App(fn, name)
    if fn is None:
        return decorate
    return decorate(fn)


class App(Command):
    """
    :class:`clik.Command` subclass that implements the :meth:`main` method.

    :meth:`main` is the user-level API for starting the application.
    """

    def __init__(self, fn, name=None):
        """
        Initialize the application object.

        :param fn: Main application generator function.
        :param name: Optional :class:`str` specifying the application name.
                     If not supplied the application name is set to
                     ``fn.__name__``.
        """
        super(App, self).__init__(Context(), fn, name=name)

    def main(self, argv=None, exit=sys.exit):
        """
        Start the application.

        :param argv: Optional list of command-line arguments. If not supplied,
                     this defaults to ``sys.argv``.
        :param exit: Optional function to call on exit. If not supplied,
                     this defaults to :func:`sys.exit`.
        :type exit: ``fn(integer_exit_code)``
        :return: Return value of ``exit``
        """
        if argv is None:  # pragma: no cover (hard to test, obviously correct)
            argv = sys.argv

        description, epilog = self._split_docstring(self._fn)
        root_parser = ArgumentParser(
            prog=self._name,
            description=description,
            epilog=epilog,
        )
        root_parser.set_defaults(**{ALLOW_UNKNOWN: False})

        magics = (args, current_app, g, parser, run_children, unknown_args)
        with self._ctx.acquire(*magics):
            with self._ctx(current_app=self, g=AttributeDict()):
                with self._ctx(args=None):
                    ec = self._configure_parser(root_parser)
                if ec:
                    return exit(ec)
                try:
                    known, unknown = root_parser.parse_known_args(argv[1:])
                except ArgumentParserExit:
                    return exit(1)
                if unknown and not getattr(known, ALLOW_UNKNOWN):
                    try:
                        root_parser.parse_args(argv[1:])
                    except ArgumentParserExit:
                        return exit(1)
                with self._ctx(args=known, unknown_args=unknown):
                    return exit(self._run())
